@@ -4,6 +4,7 @@ import (
 	"bytes"
 	"fmt"
 	"io"
+	"io/fs"
 	"io/ioutil"
 	"log"
 	"net/url"
@@ -12,7 +13,6 @@ import (
 	"path/filepath"
 	"regexp"
 	"runtime"
-	"runtime/pprof"
 	"sort"
 	"strings"
 	"time"
@@ -20,7 +20,7 @@ import (
 	"github.com/djherbis/atime"
 	humanize "github.com/dustin/go-humanize"
 	"github.com/matryer/try"
-	flag "github.com/spf13/pflag"
+	"github.com/tdewolff/argp"
 	min "github.com/tdewolff/minify/v2"
 	"github.com/tdewolff/minify/v2/css"
 	"github.com/tdewolff/minify/v2/html"
@@ -33,15 +33,18 @@ import (
 // Version is the current minify version.
 var Version = "built from source"
 
-var filetypeMime = map[string]string{
-	"css":  "text/css",
-	"htm":  "text/html",
-	"html": "text/html",
-	"js":   "application/javascript",
-	"mjs":  "application/javascript",
-	"json": "application/json",
-	"svg":  "image/svg+xml",
-	"xml":  "text/xml",
+var extMap = map[string]string{
+	"css":         "text/css",
+	"htm":         "text/html",
+	"html":        "text/html",
+	"js":          "application/javascript",
+	"mjs":         "application/javascript",
+	"json":        "application/json",
+	"rss":         "application/rss+xml",
+	"svg":         "image/svg+xml",
+	"webmanifest": "application/manifest+json",
+	"xhtml":       "application/xhtml-xml",
+	"xml":         "text/xml",
 }
 
 var (
@@ -50,7 +53,10 @@ var (
 	list               bool
 	m                  *min.M
 	matches            []string
+	matchesRegexp      []*regexp.Regexp
 	filters            []string
+	filtersRegexp      []*regexp.Regexp
+	extensions         map[string]string
 	recursive          bool
 	quiet              bool
 	verbose            int
@@ -64,30 +70,87 @@ var (
 	preserveTimestamps bool
 	preserveLinks      bool
 	mimetype           string
+	oldmimetype        string
 )
 
+type Matches struct {
+	matches *[]string
+}
+
+func (scanner Matches) Scan(s []string) (int, error) {
+	n := 0
+	for _, item := range s {
+		if strings.HasPrefix(item, "-") {
+			break
+		}
+		*scanner.matches = append(*scanner.matches, item)
+		n++
+	}
+	return n, nil
+}
+
+func (typenamer Matches) TypeName() string {
+	return "[]string"
+}
+
+type Includes struct {
+	filters *[]string
+}
+
+func (scanner Includes) Scan(s []string) (int, error) {
+	n := 0
+	for _, item := range s {
+		if strings.HasPrefix(item, "-") {
+			break
+		}
+		*scanner.filters = append(*scanner.filters, "+"+item)
+		n++
+	}
+	return n, nil
+}
+
+func (typenamer Includes) TypeName() string {
+	return "[]string"
+}
+
+type Excludes struct {
+	filters *[]string
+}
+
+func (scanner Excludes) Scan(s []string) (int, error) {
+	n := 0
+	for _, item := range s {
+		if strings.HasPrefix(item, "-") {
+			break
+		}
+		*scanner.filters = append(*scanner.filters, "-"+item)
+		n++
+	}
+	return n, nil
+}
+
+func (typenamer Excludes) TypeName() string {
+	return "[]string"
+}
+
 // Task is a minify task.
 type Task struct {
-	srcs []string
 	root string
+	srcs []string
 	dst  string
 	sync bool
 }
 
 // NewTask returns a new Task.
 func NewTask(root, input, output string, sync bool) (Task, error) {
-	if IsDir(output) {
-		root = filepath.Dir(root)
-	}
-	t := Task{[]string{input}, root, output, sync}
-	if 0 < len(output) && output[len(output)-1] == os.PathSeparator {
+	if len(output) != 0 && (output == "." || output[len(output)-1] == os.PathSeparator) {
 		rel, err := filepath.Rel(root, input)
 		if err != nil {
 			return Task{}, err
 		}
-		t.dst = filepath.Join(output, rel)
+		output = filepath.Join(output, rel)
 	}
-	return t, nil
+	return Task{root, []string{input}, output, sync}, nil
 }
 
 // Loggers.
@@ -103,11 +166,9 @@ func main() {
 }
 
 func run() int {
-	output := ""
-	filetype := ""
-	siteurl := ""
-	cpuprofile := ""
-	memprofile := ""
+	var inputs []string
+	var output string
+	var siteurl string
 
 	cssMinifier := &css.Minifier{}
 	htmlMinifier := &html.Minifier{}
@@ -116,81 +177,88 @@ func run() int {
 	svgMinifier := &svg.Minifier{}
 	xmlMinifier := &xml.Minifier{}
 
-	f := flag.NewFlagSet("minify", flag.ContinueOnError)
-	f.Usage = func() {
-		fmt.Fprintf(os.Stderr, "Usage: %s [options] [input]\n\nOptions:\n", os.Args[0])
-		f.PrintDefaults()
-		fmt.Fprintf(os.Stderr, "\nInput:\n  Files or directories, leave blank to use stdin. Specify --mime or --type to use stdin and stdout.\n")
-	}
+	f := argp.New("minify")
+	f.AddRest(&inputs, "inputs", "Input files or directories, leave blank to use stdin")
+	f.AddOpt(&output, "o", "output", nil, "Output file or directory, leave blank to use stdout")
+	f.AddOpt(&oldmimetype, "", "mime", nil, "Mimetype (eg. text/css), optional for input filenames (DEPRECATED, use --type)")
+	f.AddOpt(&mimetype, "", "type", nil, "Filetype (eg. css or text/css), optional when specifying inputs")
+	f.AddOpt(Matches{&matches}, "", "match", nil, "Filename matching pattern, only matching filenames are processed")
+	f.AddOpt(Includes{&filters}, "", "include", nil, "Path inclusion pattern, includes paths previously excluded")
+	f.AddOpt(Excludes{&filters}, "", "exclude", nil, "Path exclusion pattern, excludes paths from being processed")
+	f.AddOpt(&extensions, "", "ext", nil, "Filename extension mapping to filetype (eg. css or text/css)")
+	f.AddOpt(&recursive, "r", "recursive", false, "Recursively minify directories")
+	f.AddOpt(&hidden, "a", "all", false, "Minify all files, including hidden files and files in hidden directories")
+	f.AddOpt(&list, "l", "list", false, "List all accepted filetypes")
+	f.AddOpt(&quiet, "q", "quiet", false, "Quiet mode to suppress all output")
+	f.AddOpt(argp.Count{&verbose}, "v", "verbose", nil, "Verbose mode, set twice for more verbosity")
+	f.AddOpt(&watch, "w", "watch", false, "Watch files and minify upon changes")
+	f.AddOpt(&sync, "s", "sync", false, "Copy all files to destination directory and minify when filetype matches")
+	f.AddOpt(&preserve, "p", "preserve", []string{"mode", "ownership", "timestamps"}, "Preserve options (mode, ownership, timestamps, links, all)")
+	f.AddOpt(&bundle, "b", "bundle", false, "Bundle files by concatenation into a single file")
+	f.AddOpt(&version, "", "version", false, "Version")
 
-	f.BoolVarP(&help, "help", "h", false, "Show usage")
-	f.StringVarP(&output, "output", "o", "", "Output file or directory (must have trailing slash), leave blank to use stdout")
-	f.StringVar(&mimetype, "mime", "", "Mimetype (eg. text/css), optional for input filenames, has precedence over --type")
-	f.StringVar(&filetype, "type", "", "Filetype (eg. css), optional for input filenames")
-	f.String("match", "", "Filename matching pattern, only matching files are processed")
-	f.String("include", "", "Filename inclusion pattern, includes files previously excluded")
-	f.String("exclude", "", "Filename exclusion pattern, excludes files from being processed")
-	f.BoolVarP(&recursive, "recursive", "r", false, "Recursively minify directories")
-	f.BoolVarP(&hidden, "all", "a", false, "Minify all files, including hidden files and files in hidden directories")
-	f.BoolVarP(&list, "list", "l", false, "List all accepted filetypes")
-	f.BoolVarP(&quiet, "quiet", "q", false, "Quiet mode to suppress all output")
-	f.CountVarP(&verbose, "verbose", "v", "Verbose mode, set twice for more verbosity")
-	f.BoolVarP(&watch, "watch", "w", false, "Watch files and minify upon changes")
-	f.BoolVarP(&sync, "sync", "s", false, "Copy all files to destination directory and minify when filetype matches")
-	f.StringSliceVarP(&preserve, "preserve", "p", nil, "Preserve options (mode, ownership, timestamps, links, all)")
-	f.Lookup("preserve").NoOptDefVal = "mode,ownership,timestamps"
-	f.BoolVarP(&bundle, "bundle", "b", false, "Bundle files by concatenation into a single file")
-	f.BoolVar(&version, "version", false, "Version")
+	f.AddOpt(&siteurl, "", "url", nil, "URL of file to enable URL minification")
+	f.AddOpt(&cssMinifier.Precision, "", "css-precision", 0, "Number of significant digits to preserve in numbers, 0 is all")
+	f.AddOpt(&htmlMinifier.KeepComments, "", "html-keep-comments", false, "Preserve all comments")
+	f.AddOpt(&htmlMinifier.KeepConditionalComments, "", "html-keep-conditional-comments", false, "Preserve all IE conditional comments")
+	f.AddOpt(&htmlMinifier.KeepDefaultAttrVals, "", "html-keep-default-attrvals", false, "Preserve default attribute values")
+	f.AddOpt(&htmlMinifier.KeepDocumentTags, "", "html-keep-document-tags", false, "Preserve html, head and body tags")
+	f.AddOpt(&htmlMinifier.KeepEndTags, "", "html-keep-end-tags", false, "Preserve all end tags")
+	f.AddOpt(&htmlMinifier.KeepWhitespace, "", "html-keep-whitespace", false, "Preserve whitespace characters but still collapse multiple into one")
+	f.AddOpt(&htmlMinifier.KeepQuotes, "", "html-keep-quotes", false, "Preserve quotes around attribute values")
+	f.AddOpt(&jsMinifier.Precision, "", "js-precision", 0, "Number of significant digits to preserve in numbers, 0 is all")
+	f.AddOpt(&jsMinifier.KeepVarNames, "", "js-keep-var-names", false, "Preserve original variable names")
+	f.AddOpt(&jsMinifier.Version, "", "js-version", 0, "ECMAScript version to toggle supported optimizations (e.g. 2019, 2020), by default 0 is the latest version")
+	f.AddOpt(&jsonMinifier.Precision, "", "json-precision", 0, "Number of significant digits to preserve in numbers, 0 is all")
+	f.AddOpt(&jsonMinifier.KeepNumbers, "", "json-keep-numbers", false, "Preserve original numbers instead of minifying them")
+	f.AddOpt(&svgMinifier.KeepComments, "", "svg-keep-comments", false, "Preserve all comments")
+	f.AddOpt(&svgMinifier.Precision, "", "svg-precision", 0, "Number of significant digits to preserve in numbers, 0 is all")
+	f.AddOpt(&xmlMinifier.KeepWhitespace, "", "xml-keep-whitespace", false, "Preserve whitespace characters but still collapse multiple into one")
+	f.Parse()
 
-	f.StringVar(&siteurl, "url", "", "URL of file to enable URL minification")
-	f.StringVar(&cpuprofile, "cpuprofile", "", "Export CPU profile")
-	f.StringVar(&memprofile, "memprofile", "", "Export memory profile")
-	f.IntVar(&cssMinifier.Precision, "css-precision", 0, "Number of significant digits to preserve in numbers, 0 is all")
-	f.BoolVar(&htmlMinifier.KeepComments, "html-keep-comments", false, "Preserve all comments")
-	f.BoolVar(&htmlMinifier.KeepConditionalComments, "html-keep-conditional-comments", false, "Preserve all IE conditional comments")
-	f.BoolVar(&htmlMinifier.KeepDefaultAttrVals, "html-keep-default-attrvals", false, "Preserve default attribute values")
-	f.BoolVar(&htmlMinifier.KeepDocumentTags, "html-keep-document-tags", false, "Preserve html, head and body tags")
-	f.BoolVar(&htmlMinifier.KeepEndTags, "html-keep-end-tags", false, "Preserve all end tags")
-	f.BoolVar(&htmlMinifier.KeepWhitespace, "html-keep-whitespace", false, "Preserve whitespace characters but still collapse multiple into one")
-	f.BoolVar(&htmlMinifier.KeepQuotes, "html-keep-quotes", false, "Preserve quotes around attribute values")
-	f.IntVar(&jsMinifier.Precision, "js-precision", 0, "Number of significant digits to preserve in numbers, 0 is all")
-	f.BoolVar(&jsMinifier.KeepVarNames, "js-keep-var-names", false, "Preserve original variable names")
-	f.IntVar(&jsMinifier.Version, "js-version", 0, "ECMAScript version to toggle supported optimizations (e.g. 2019, 2020), by default 0 is the latest version")
-	f.IntVar(&jsonMinifier.Precision, "json-precision", 0, "Number of significant digits to preserve in numbers, 0 is all")
-	f.BoolVar(&jsonMinifier.KeepNumbers, "json-keep-numbers", false, "Preserve original numbers instead of minifying them")
-	f.BoolVar(&svgMinifier.KeepComments, "svg-keep-comments", false, "Preserve all comments")
-	f.IntVar(&svgMinifier.Precision, "svg-precision", 0, "Number of significant digits to preserve in numbers, 0 is all")
-	f.BoolVar(&xmlMinifier.KeepWhitespace, "xml-keep-whitespace", false, "Preserve whitespace characters but still collapse multiple into one")
-	if len(os.Args) == 1 {
+	if version {
 		if !quiet {
-			fmt.Printf("minify: must specify --mime or --type in order to use stdin and stdout\n")
-			fmt.Printf("Try 'minify --help' for more information\n")
+			fmt.Printf("minify %s\n", Version)
 		}
-		return 1
+		return 0
 	}
-	err := f.ParseAll(os.Args[1:], func(flag *flag.Flag, value string) error {
-		if flag.Name == "match" || flag.Name == "include" || flag.Name == "exclude" {
-			for _, filter := range strings.Split(value, ",") {
-				if flag.Name == "match" {
-					matches = append(matches, filter)
-				} else if flag.Name == "include" {
-					filters = append(filters, "+"+filter)
-				} else {
-					filters = append(filters, "-"+filter)
+
+	for ext, filetype := range extensions {
+		if mimetype, ok := extMap[filetype]; ok {
+			filetype = mimetype
+		}
+		extMap[ext] = filetype
+	}
+
+	if list {
+		if !quiet {
+			n := 0
+			var keys []string
+			for k := range extMap {
+				keys = append(keys, k)
+				if n < len(k) {
+					n = len(k)
 				}
 			}
-			return nil
+			sort.Strings(keys)
+			for _, k := range keys {
+				fmt.Println(k + strings.Repeat(" ", n-len(k)+2) + extMap[k])
+			}
 		}
-		return f.Set(flag.Name, value)
-	})
-	if err != nil {
+		return 0
+	}
+
+	if len(inputs) == 0 && mimetype == "" && oldmimetype == "" {
 		if !quiet {
-			fmt.Printf("minify: %v\n", err)
+			fmt.Printf("minify: must specify --type in order to use stdin and stdout\n")
 			fmt.Printf("Try 'minify --help' for more information\n")
 		}
 		return 1
+	} else if len(inputs) == 1 && inputs[0] == "-" {
+		inputs = inputs[:0]
+	} else if output == "-" {
+		output = ""
 	}
-	inputs := f.Args()
 	useStdin := len(inputs) == 0
 
 	Error = log.New(ioutil.Discard, "", 0)
@@ -206,76 +274,21 @@ func run() int {
 		}
 	}
 
-	if help {
-		f.Usage()
-		return 0
-	}
-
-	if version {
-		if !quiet {
-			fmt.Printf("minify %s\n", Version)
-		}
-		return 0
-	}
-
-	if list {
-		if !quiet {
-			var keys []string
-			for k := range filetypeMime {
-				keys = append(keys, k)
-			}
-			sort.Strings(keys)
-			for _, k := range keys {
-				fmt.Println(k + "\t" + filetypeMime[k])
-			}
-		}
-		return 0
-	}
-
-	if cpuprofile != "" {
-		f, err := os.Create(cpuprofile)
-		if err != nil {
-			Error.Println(err)
-			return 1
-		}
-		if err = pprof.StartCPUProfile(f); err != nil {
-			Error.Println(err)
-			return 1
-		}
-		defer func() {
-			pprof.StopCPUProfile()
-			if err = f.Close(); err != nil {
-				Error.Println(err)
-			}
-		}()
-	}
-
-	if memprofile != "" {
-		defer func() {
-			f, err := os.Create(memprofile)
-			if err != nil {
-				Error.Println(err)
-			}
-			if err = pprof.WriteHeapProfile(f); err != nil {
-				Error.Println(err)
-			}
-			if err = f.Close(); err != nil {
-				Error.Println(err)
-			}
-		}()
-	}
-
+	// compile matches and regexps
+	var err error
 	if 0 < len(matches) {
-		for _, filter := range matches {
-			if _, err = regexp.Compile(filter); err != nil {
+		matchesRegexp = make([]*regexp.Regexp, len(matches))
+		for i, pattern := range matches {
+			if matchesRegexp[i], err = compilePattern(pattern); err != nil {
 				Error.Println(err)
 				return 1
 			}
 		}
 	}
 	if 0 < len(filters) {
-		for _, filter := range filters {
-			if _, err = regexp.Compile(filter[1:]); err != nil {
+		filtersRegexp = make([]*regexp.Regexp, len(filters))
+		for i, pattern := range filters {
+			if filtersRegexp[i], err = compilePattern(pattern[1:]); err != nil {
 				Error.Println(err)
 				return 1
 			}
@@ -283,31 +296,43 @@ func run() int {
 	}
 
 	// detect mimetype, mimetype=="" means we'll infer mimetype from file extensions
-	if mimetype == "" && filetype != "" {
+	if oldmimetype != "" {
+		Error.Printf("deprecated use of '--mime %v', please use '--type %v' instead", oldmimetype, oldmimetype)
+		mimetype = oldmimetype
+	}
+	if slash := strings.Index(mimetype, "/"); slash == -1 && 0 < len(mimetype) {
 		var ok bool
-		if mimetype, ok = filetypeMime[filetype]; !ok {
-			Error.Println("cannot find mimetype for filetype", filetype)
+		if mimetype, ok = extMap[mimetype]; !ok {
+			Error.Println("unknown filetype", mimetype)
 			return 1
 		}
 	}
 
-	if (useStdin || output == "") && (watch || sync || recursive) {
+	if (useStdin || output == "") && (watch || sync) {
 		if watch {
-			Error.Println("--watch doesn't work on stdin and stdout, specify input and output")
+			Error.Println("--watch doesn't work with stdin and stdout, specify input and output")
 		}
 		if sync {
-			Error.Println("--sync doesn't work on stdin and stdout, specify input and output")
+			Error.Println("--sync doesn't work with stdin and stdout, specify input and output")
+		}
+		return 1
+	} else if useStdin && (bundle || recursive) {
+		if bundle {
+			Error.Println("--bundle doesn't work with stdin, specify input")
 		}
 		if recursive {
-			Error.Println("--recursive doesn't work on stdin and stdout, specify input and output")
+			Error.Println("--recursive doesn't work with stdin, specify input")
 		}
 		return 1
+	} else if output == "" && recursive && !bundle {
+		Error.Println("--recursive doesn't work with stdout, specify output or use --bundle")
+		return 1
 	}
 	if mimetype == "" && useStdin {
-		Error.Println("must specify --mime or --type for stdin")
+		Error.Println("must specify --type for stdin")
 		return 1
 	} else if mimetype != "" && sync {
-		Error.Println("must specify either --sync or --mime/--type")
+		Error.Println("must specify either --sync or --type")
 		return 1
 	}
 	if mimetype == "" {
@@ -315,7 +340,7 @@ func run() int {
 	} else {
 		Info.Println("use mimetype", mimetype)
 	}
-	if len(preserve) != 0 {
+	if f.IsSet("preserve") {
 		if bundle {
 			Error.Println("--preserve cannot be used together with --bundle")
 			return 1
@@ -345,7 +370,14 @@ func run() int {
 	////////////////
 
 	for i, input := range inputs {
+		if input == "-" {
+			Error.Println("cannot mix files and stdin as input")
+			return 1
+		}
 		inputs[i] = filepath.Clean(input)
+		if input[len(input)-1] == os.PathSeparator {
+			inputs[i] += string(os.PathSeparator)
+		}
 	}
 
 	// set output file or directory, empty means stdout
@@ -379,7 +411,7 @@ func run() int {
 		Info.Println("minify to stdout")
 	} else if !dirDst {
 		Info.Println("minify to output file", output)
-	} else if output == "./" {
+	} else if output == "."+string(os.PathSeparator) {
 		Info.Println("minify to current working directory")
 	} else {
 		Info.Println("minify to output directory", output)
@@ -399,7 +431,8 @@ func run() int {
 		tasks = append(tasks, task)
 		roots = append(roots, "")
 	} else {
-		tasks, roots, err = createTasks(inputs, output)
+		fsys := NewFS()
+		tasks, roots, err = createTasks(fsys, inputs, output)
 		if err != nil {
 			Error.Println(err)
 			return 1
@@ -541,12 +574,27 @@ func minifyWorker(chanTasks <-chan Task, chanFails chan<- int) {
 	chanFails <- fails
 }
 
+// compilePattern returns *regexp.Regexp or glob.Glob
+func compilePattern(pattern string) (*regexp.Regexp, error) {
+	if len(pattern) == 0 || pattern[0] != '~' {
+		if strings.HasPrefix(pattern, `\~`) {
+			pattern = pattern[1:]
+		}
+		pattern = regexp.QuoteMeta(pattern)
+		pattern = strings.ReplaceAll(pattern, `\*\*`, `.*`)
+		pattern = strings.ReplaceAll(pattern, `\*`, fmt.Sprintf(`[^%c]*`, filepath.Separator))
+		pattern = strings.ReplaceAll(pattern, `\?`, fmt.Sprintf(`[^%c]?`, filepath.Separator))
+		pattern = "^" + pattern + "$"
+	}
+	return regexp.Compile(pattern)
+}
+
 func fileFilter(filename string) bool {
 	if 0 < len(matches) {
 		match := false
-		for _, filter := range matches {
-			pattern := regexp.MustCompile(filter)
-			if pattern.MatchString(filename) {
+		base := filepath.Base(filename)
+		for _, re := range matchesRegexp {
+			if re.MatchString(base) {
 				match = true
 				break
 			}
@@ -555,13 +603,13 @@ func fileFilter(filename string) bool {
 			return false
 		}
 	}
-	for _, filter := range filters {
-		pattern := regexp.MustCompile(filter[1:])
-		if pattern.MatchString(filename) {
-			return filter[0] == '+'
+	match := true
+	for i, re := range filtersRegexp {
+		if re.MatchString(filename) {
+			match = filters[i][0] == '+'
 		}
 	}
-	return true
+	return match
 }
 
 func fileMatches(filename string) bool {
@@ -575,21 +623,24 @@ func fileMatches(filename string) bool {
 	if 0 < len(ext) {
 		ext = ext[1:]
 	}
-	if _, ok := filetypeMime[ext]; !ok {
+	if _, ok := extMap[ext]; !ok {
 		return false
 	}
 	return true
 }
 
-func createTasks(inputs []string, output string) ([]Task, []string, error) {
+func createTasks(fsys fs.FS, inputs []string, output string) ([]Task, []string, error) {
 	tasks := []Task{}
 	roots := []string{}
 	for _, input := range inputs {
-		var info os.FileInfo
+		root := filepath.Clean(filepath.Dir(input))
+		input = filepath.Clean(input)
+
 		var err error
+		var info os.FileInfo
 		if !preserveLinks {
 			// follow and dereference symlinks
-			info, err = os.Stat(input)
+			info, err = fs.Stat(fsys, input)
 		} else {
 			info, err = os.Lstat(input)
 		}
@@ -597,7 +648,6 @@ func createTasks(inputs []string, output string) ([]Task, []string, error) {
 			return nil, nil, err
 		}
 
-		root := filepath.Clean(filepath.Dir(input) + string(os.PathSeparator) + ".")
 		if preserveLinks && info.Mode()&os.ModeSymlink != 0 {
 			// copy symlink as is
 			if !sync {
@@ -610,7 +660,7 @@ func createTasks(inputs []string, output string) ([]Task, []string, error) {
 			}
 			tasks = append(tasks, task)
 		} else if info.Mode().IsRegular() {
-			valid := fileFilter(info.Name()) // don't filter mimetype
+			valid := fileFilter(input) // don't filter mimetype
 			if valid || sync {
 				task, err := NewTask(root, input, output, !valid)
 				if err != nil {
@@ -623,47 +673,36 @@ func createTasks(inputs []string, output string) ([]Task, []string, error) {
 				Warning.Println("--recursive not specified, omitting directory", input)
 				continue
 			}
-			root = input
-			roots = append(roots, root)
 
-			var walkFn func(string, DirEntry, error) error
-			walkFn = func(path string, d DirEntry, err error) error {
+			var walkFn func(string, fs.DirEntry, error) error
+			walkFn = func(input string, d fs.DirEntry, err error) error {
 				if err != nil {
 					return err
 				} else if d.Name() == "." || d.Name() == ".." {
 					return nil
 				} else if d.Name() == "" || !hidden && d.Name()[0] == '.' {
 					if d.IsDir() {
-						return SkipDir
+						return fs.SkipDir
 					}
 					return nil
 				}
 
-				input := path
-				if filepath.IsAbs(input) {
-					// path is absolute, clean it
-					input = filepath.Clean(input)
-				} else {
-					// path is relative, make an absolute path - Join() returns a clean result
-					input = filepath.Join(root, input)
-				}
-
 				if !preserveLinks && d.Type()&os.ModeSymlink != 0 {
 					// follow and dereference symlinks
-					info, err := os.Stat(input)
+					info, err := fs.Stat(fsys, input)
 					if err != nil {
 						return err
 					}
 					if info.IsDir() {
-						return WalkDir(DirFS(root), path, walkFn)
+						return fs.WalkDir(fsys, input, walkFn)
 					}
-					d = &statDirEntry{info}
+					d = fs.FileInfoToDirEntry(info)
 				}
 
 				if preserveLinks && d.Type()&os.ModeSymlink != 0 {
 					// copy symlink as is
 					if !sync {
-						Warning.Println("--sync not specified, omitting symbolic link", path)
+						Warning.Println("--sync not specified, omitting symbolic link", input)
 						return nil
 					}
 					task, err := NewTask(root, input, output, true)
@@ -672,7 +711,7 @@ func createTasks(inputs []string, output string) ([]Task, []string, error) {
 					}
 					tasks = append(tasks, task)
 				} else if d.Type().IsRegular() {
-					valid := fileMatches(d.Name())
+					valid := fileMatches(input)
 					if valid || sync {
 						task, err := NewTask(root, input, output, !valid)
 						if err != nil {
@@ -683,9 +722,10 @@ func createTasks(inputs []string, output string) ([]Task, []string, error) {
 				}
 				return nil
 			}
-			if err := WalkDir(DirFS(root), ".", walkFn); err != nil {
+			if err := fs.WalkDir(fsys, input, walkFn); err != nil {
 				return nil, nil, err
 			}
+			roots = append(roots, root)
 		} else {
 			return nil, nil, fmt.Errorf("not a file or directory %s", input)
 		}
@@ -693,62 +733,6 @@ func createTasks(inputs []string, output string) ([]Task, []string, error) {
 	return tasks, roots, nil
 }
 
-func openInputFile(input string) (io.ReadCloser, error) {
-	var r *os.File
-	if input == "" {
-		r = os.Stdin
-	} else {
-		err := try.Do(func(attempt int) (bool, error) {
-			var ferr error
-			r, ferr = os.Open(input)
-			return attempt < 5, ferr
-		})
-
-		if err != nil {
-			return nil, fmt.Errorf("open input file %q: %w", input, err)
-		}
-	}
-	return r, nil
-}
-
-func openOutputFile(output string) (*os.File, error) {
-	var w *os.File
-	if output == "" {
-		w = os.Stdout
-	} else {
-		dir := filepath.Dir(output)
-		if err := os.MkdirAll(dir, 0777); err != nil {
-			return nil, fmt.Errorf("creating directory %q: %w", dir, err)
-		}
-
-		err := try.Do(func(attempt int) (bool, error) {
-			var ferr error
-			w, ferr = os.OpenFile(output, os.O_WRONLY|os.O_TRUNC|os.O_CREATE, 0666)
-			return attempt < 5, ferr
-		})
-
-		if err != nil {
-			return nil, fmt.Errorf("open output file %q: %w", output, err)
-		}
-	}
-	return w, nil
-}
-
-func createSymlink(input, output string) error {
-	if _, err := os.Stat(output); err == nil {
-		if err = os.Remove(output); err != nil {
-			return err
-		}
-	}
-	if err := os.MkdirAll(filepath.Dir(output), 0777); err != nil {
-		return err
-	}
-	if err := os.Symlink(input, output); err != nil {
-		return err
-	}
-	return nil
-}
-
 func minify(t Task) bool {
 	// synchronizing files that are not minified but just copied to the same directory, no action needed
 	if t.sync {
@@ -775,15 +759,15 @@ func minify(t Task) bool {
 			if 0 < len(ext) {
 				ext = ext[1:]
 			}
-			srcMimetype, ok := filetypeMime[ext]
+			srcMimetype, ok := extMap[ext]
 			if !ok {
-				Warning.Println("cannot infer mimetype from extension in", src, ", set --type or --mime explicitly")
+				Warning.Println("cannot infer mimetype from extension in", src, ", set --type explicitly")
 				return false
 			}
 			if fileMimetype == "" {
 				fileMimetype = srcMimetype
 			} else if srcMimetype != fileMimetype {
-				Warning.Println("inferred mimetype", srcMimetype, "of", src, "for concatenation unequal to previous mimetypes, set --type or --mime explicitly")
+				Warning.Println("inferred mimetype", srcMimetype, "of", src, "for concatenation unequal to previous mimetypes, set --type explicitly")
 				return false
 			}
 		}
@@ -823,10 +807,11 @@ func minify(t Task) bool {
 	if len(t.srcs) == 1 {
 		fr, err = openInputFile(t.srcs[0])
 	} else {
-		fr, err = newConcatFileReader(t.srcs, openInputFile)
-		if err == nil && fileMimetype == filetypeMime["js"] {
-			fr.(*concatFileReader).SetSeparator([]byte(";"))
+		var sep []byte
+		if err == nil && fileMimetype == extMap["js"] {
+			sep = []byte(";\n")
 		}
+		fr, err = openInputFiles(t.srcs, sep)
 	}
 	if err != nil {
 		Error.Println(err)
@@ -960,25 +945,3 @@ Next:
 		goto Next
 	}
 }
-
-// IsDir returns true if the passed string looks like it specifies a directory, false otherwise.
-func IsDir(dir string) bool {
-	info, err := os.Lstat(dir)
-	return err == nil && info.Mode().IsDir() && info.Mode()&os.ModeSymlink == 0
-}
-
-// SameFile returns true if the two file paths specify the same path.
-// While Linux is case-preserving case-sensitive (and therefore a string comparison will work),
-// Windows is case-preserving case-insensitive; we use os.SameFile() to work cross-platform.
-func SameFile(filename1 string, filename2 string) (bool, error) {
-	fi1, err := os.Stat(filename1)
-	if err != nil {
-		return false, err
-	}
-
-	fi2, err := os.Stat(filename2)
-	if err != nil {
-		return false, err
-	}
-	return os.SameFile(fi1, fi2), nil
-}
